# -*- coding: utf-8 -*-

import unittest
import torch as th

from leibniz.nn.net import resunet
from leibniz.nn.layer.hyperbolic import HyperBasic
from leibniz.nn.layer.hyperbolic import HyperBottleneck
from leibniz.nn.layer.senet import SEBasicBlock, SEBottleneck
from leibniz.nn.layer.hyperbolic2 import (
    HyperBasic as HyperBasic2,
    HyperBottleneck as HyperBottleneck2,
)


class TestUnet(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test1D(self):
        net = resunet(1, 1, spatial=(32,))
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="instance")
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer")
        net(th.rand(1, 1, 16))

    def test2D(self):
        resunet(1, 1, spatial=(16, 16))
        resunet(1, 1, spatial=(16, 32))
        resunet(1, 1, spatial=(32, 16))
        net = resunet(
            1, 1, spatial=(32, 16), scales=[[0, -1], [0, -1], [0, -1], [0, -1]]
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
        )
        net(th.rand(1, 1, 32, 16))

    def test3D(self):
        resunet(1, 1, spatial=(16, 16, 16))
        resunet(1, 1, spatial=(32, 16, 16))
        resunet(1, 1, spatial=(16, 32, 16))
        resunet(1, 1, spatial=(16, 16, 32))
        resunet(1, 1, spatial=(11, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
        )
        net(th.rand(1, 1, 4, 16, 32))

    def testHyp1D(self):
        net = resunet(1, 1, spatial=(32,), block=HyperBasic)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="instance", block=HyperBasic)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer", block=HyperBasic)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), block=HyperBottleneck)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="instance", block=HyperBottleneck)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer", block=HyperBottleneck)
        net(th.rand(1, 1, 16))

    def testHyp2D(self):
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            block=HyperBasic,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 32, 16))

    def testHyp3D(self):
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            block=HyperBasic,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 4, 16, 32))

    def testSE1D(self):
        net = resunet(1, 1, spatial=(32,), block=SEBasicBlock)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="instance", block=SEBasicBlock)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer", block=SEBasicBlock)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), block=SEBottleneck)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="instance", block=SEBottleneck)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer", block=SEBottleneck)
        net(th.rand(1, 1, 16))

    def testSE2D(self):
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            block=SEBasicBlock,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
            block=SEBasicBlock,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
            block=SEBasicBlock,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            block=SEBottleneck,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
            block=SEBottleneck,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
            block=SEBottleneck,
        )
        net(th.rand(1, 1, 32, 16))

    def testSE3D(self):
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            block=SEBasicBlock,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
            block=SEBasicBlock,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
            block=SEBasicBlock,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            block=SEBottleneck,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
            block=SEBottleneck,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
            block=SEBottleneck,
        )
        net(th.rand(1, 1, 4, 16, 32))

    def testHyp2DGroupNorm(self):
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="group",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="group",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="group",
            block=HyperBasic,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="group",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="group",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="group",
            block=HyperBottleneck,
        )
        net(th.rand(1, 1, 32, 16))

    def testHyp1D2(self):
        net = resunet(1, 1, spatial=(32,), block=HyperBasic2)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="instance", block=HyperBasic2)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer", block=HyperBasic2)
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), block=HyperBottleneck2)
        net(th.rand(1, 1, 16))
        net = resunet(
            1, 1, spatial=(32,), normalizor="instance", block=HyperBottleneck2
        )
        net(th.rand(1, 1, 16))
        net = resunet(1, 1, spatial=(32,), normalizor="layer", block=HyperBottleneck2)
        net(th.rand(1, 1, 16))

    def testHyp2D2(self):
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            block=HyperBasic2,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
            block=HyperBasic2,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
            block=HyperBasic2,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            block=HyperBottleneck2,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
            block=HyperBottleneck2,
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
            block=HyperBottleneck2,
        )
        net(th.rand(1, 1, 32, 16))

    def testHyp3D2(self):
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            block=HyperBasic2,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
            block=HyperBasic2,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
            block=HyperBasic2,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            block=HyperBottleneck2,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="instance",
            block=HyperBottleneck2,
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resunet(
            1,
            1,
            spatial=(4, 16, 32),
            scales=[[0, -1, -1], [-1, -1, -1], [0, -1, -1], [-1, -1, -1]],
            normalizor="layer",
            block=HyperBottleneck2,
        )
        net(th.rand(1, 1, 4, 16, 32))



# -*- coding: utf-8 -*-

import unittest

import numpy as np
import leibniz as lbnz

from leibniz.core3d.vec3 import box
from leibniz.core3d.gridsys.regular3 import RegularGrid


class TestFrame(unittest.TestCase):

    def setUp(self):
        lbnz.bind(
            RegularGrid(
                basis="lng,lat,alt",
                W=51,
                L=51,
                H=51,
                east=119.0,
                west=114.0,
                north=42.3,
                south=37.3,
                upper=16000.0,
                lower=0.0,
            )
        )
        lbnz.use("thetaphir")
        lbnz.use("xyz")
        lbnz.use("x,y,z")

    def tearDown(self):
        lbnz.clear()

    def test_basis_ortho(self):
        phx, phy, phz = lbnz.thetaphir.phi
        phx = phx.cpu().numpy()
        phy = phy.cpu().numpy()
        phz = phz.cpu().numpy()
        self.assertAlmostEqual(1, np.max(phx * phx + phy * phy + phz * phz))
        self.assertAlmostEqual(1, np.min(phx * phx + phy * phy + phz * phz))

        thx, thy, thz = lbnz.thetaphir.theta
        thx = thx.cpu().numpy()
        thy = thy.cpu().numpy()
        thz = thz.cpu().numpy()
        val = thx * thx + thy * thy + thz * thz
        self.assertAlmostEqual(1, np.max(val))
        self.assertAlmostEqual(1, np.min(val))

        rx, ry, rz = lbnz.thetaphir.r
        rx = rx.cpu().numpy()
        ry = ry.cpu().numpy()
        rz = rz.cpu().numpy()
        self.assertAlmostEqual(1, np.max(rx * rx + ry * ry + rz * rz))
        self.assertAlmostEqual(1, np.min(rx * rx + ry * ry + rz * rz))

        self.assertAlmostEqual(0, np.max(phx * thx + phy * thy + phz * thz))
        self.assertAlmostEqual(0, np.min(phx * rx + phy * ry + phz * rz))
        self.assertAlmostEqual(0, np.min(thx * rx + thy * ry + thz * rz))

    def test_basis_righthand(self):
        d = (
            box(lbnz.thetaphir.theta, lbnz.thetaphir.phi, lbnz.thetaphir.r)
            .cpu()
            .numpy()
        )
        self.assertAlmostEqual(1, np.mean(d))
        self.assertAlmostEqual(1, np.min(d))
        self.assertAlmostEqual(1, np.max(d))

        d = box(lbnz.xyz.x, lbnz.xyz.y, lbnz.xyz.z).cpu().numpy()
        self.assertAlmostEqual(1, np.mean(d))
        self.assertAlmostEqual(1, np.min(d))
        self.assertAlmostEqual(1, np.max(d))

    def test_basis_diff(self):
        from leibniz.core3d.basis.theta_phi_r import transform

        x, y, z = transform(lbnz.theta + 0.0001, lbnz.phi, lbnz.r)
        thx, thy, thz = lbnz.normalize((x - lbnz.x, y - lbnz.y, z - lbnz.z))
        a = lbnz.dot((thx, thy, thz), lbnz.thetaphir.theta)
        self.assertAlmostEqual(1, np.mean(a.cpu().numpy()))

        x, y, z = transform(lbnz.theta, lbnz.phi + 0.0001, lbnz.r)
        phx, phy, phz = lbnz.normalize((x - lbnz.x, y - lbnz.y, z - lbnz.z))
        a = lbnz.dot((phx, phy, phz), lbnz.thetaphir.phi)
        self.assertAlmostEqual(1, np.mean(a.cpu().numpy()))

        x, y, z = transform(lbnz.theta, lbnz.phi, lbnz.r + 0.0001)
        rx, ry, rz = lbnz.normalize((x - lbnz.x, y - lbnz.y, z - lbnz.z))
        a = lbnz.dot((rx, ry, rz), lbnz.thetaphir.r)
        self.assertAlmostEqual(1, np.mean(a.cpu().numpy()))



# -*- coding: utf-8 -*-

import unittest
import torch as th

from leibniz.nn.net import resunet


class TestUnet(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test2D(self):
        resunet(1, 1, spatial=(16, 16))
        resunet(1, 1, spatial=(16, 32))
        resunet(1, 1, spatial=(32, 16))
        net = resunet(
            1, 1, spatial=(32, 16), scales=[[0, -1], [0, -1], [0, -1], [0, -1]]
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
        )
        net(th.rand(1, 1, 32, 16))
        net = resunet(
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
        )
        net(th.rand(1, 1, 32, 16))



# -*- coding: utf-8 -*-


_name_ = "x,y,z"
_params_ = ("x", "y", "z")


def transform(x, y, z, **kwargs):
    return x, y, z


def dtransform(x, y, z, dx, dy, dz, **kwargs):
    return dx, dy, dz


def inverse(x, y, z, **kwargs):
    return x, y, z


def dinverse(x, y, z, dx, dy, dz, **kwargs):
    return dx, dy, dz



# -*- coding: utf-8 -*-

import unittest

import numpy as np
import torch as th
import leibniz as lbnz

from torch import Tensor
from leibniz.core3d.gridsys.regular3 import RegularGrid


class TestOpsXYZ(unittest.TestCase):

    def setUp(self):
        lbnz.bind(
            RegularGrid(
                basis="x,y,z",
                L=51,
                W=51,
                H=51,
                east=6.0,
                west=1.0,
                north=6.0,
                south=1.0,
                upper=6.0,
                lower=1.0,
            )
        )
        lbnz.use("xyz")
        lbnz.use("theta,phi,r")
        lbnz.use("thetaphir")

    def tearDown(self):
        lbnz.clear()

    def magnitude(self, var):
        # return magnitude of variable to test accuracy
        if th.is_tensor(var):
            var = var.cpu().numpy()
        if isinstance(var, np.ndarray):
            return 10 ** int(np.log10(np.abs(var.max())) + 1)
        else:
            return 10 ** int(np.log10(np.abs(var)) + 1)

    def assertAlmostEqualWithMagnitude(self, excepted, test, magnitude=6):
        if (
            isinstance(excepted, Tensor) and th.abs(excepted).max().cpu().numpy() == 0.0
        ) or (isinstance(excepted, float) and excepted == 0.0):
            return self.assertAlmostEqual(0.0, test.max().cpu().numpy(), magnitude)
        else:
            mag = self.magnitude(excepted)
            err = th.abs(excepted - test) / mag
            return self.assertAlmostEqual(0.0, err.max().cpu().numpy(), magnitude)

    def assertAlmostEqualWithMagExceptBoundary(self, excepted, test, magnitude=6):
        if isinstance(excepted, float):
            test = test[:, :, 1:-1, 1:-1, 1:-1]
            return self.assertAlmostEqualWithMagnitude(excepted, test, magnitude)
        else:
            excepted, test = (
                excepted[:, :, 1:-1, 1:-1, 1:-1],
                test[:, :, 1:-1, 1:-1, 1:-1],
            )
            return self.assertAlmostEqualWithMagnitude(excepted, test, magnitude)

    def test_grad0(self):
        dLx, dLy, dLz = lbnz.dL

        g0, g1, g2 = lbnz.grad(lbnz.x)

        self.assertAlmostEqualWithMagExceptBoundary(5.0, g0 * dLx * (lbnz.W - 1), 6)
        self.assertAlmostEqualWithMagExceptBoundary(0.0, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(0.0, g2, 6)

        g0, g1, g2 = lbnz.grad(lbnz.y)

        self.assertAlmostEqualWithMagExceptBoundary(0.0, g0, 6)
        self.assertAlmostEqualWithMagExceptBoundary(5.0, g1 * dLy * (lbnz.L - 1), 6)
        self.assertAlmostEqualWithMagExceptBoundary(0.0, g2, 6)

        g0, g1, g2 = lbnz.grad(lbnz.z)

        self.assertAlmostEqualWithMagExceptBoundary(0.0, g0, 6)
        self.assertAlmostEqualWithMagExceptBoundary(0.0, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(5.0, g2 * dLz * (lbnz.H - 1), 6)

    def test_grad1(self):
        # f = x * y * y + z * z * z
        # expect that grad_f = y * y, 2 * x * y, 3 * z * z

        fld = lbnz.x * lbnz.y * lbnz.y + lbnz.z * lbnz.z * lbnz.z
        e0, e1, e2 = lbnz.y * lbnz.y, 2 * lbnz.x * lbnz.y, 3 * lbnz.z * lbnz.z
        g0, g1, g2 = lbnz.grad(fld)

        self.assertAlmostEqualWithMagExceptBoundary(e0, g0, 3)
        self.assertAlmostEqualWithMagExceptBoundary(e1, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(e2, g2, 4)

    def test_div1(self):
        # F = (0, phi * theta, 0);
        # expect that div_F = theta * (-z * y / (r ** 2 * sqrt(r ** 2 - z ** 2))) + phi * x / (x ** 2 + y ** 2)

        fld = lbnz.zero, lbnz.phi * lbnz.theta, lbnz.zero
        expected = lbnz.theta * (
            -lbnz.z * lbnz.y / (lbnz.r**2 * th.sqrt(lbnz.r**2 - lbnz.z**2))
        ) + lbnz.phi * lbnz.x / (lbnz.x**2 + lbnz.y**2)
        test = lbnz.div(fld)

        self.assertAlmostEqualWithMagExceptBoundary(expected, test, 2)

    def test_div2(self):
        # F = (-y, x * y, z);
        # expect that div_F = x + 1

        fld = -lbnz.y, lbnz.x * lbnz.y, lbnz.z
        expected = lbnz.x + 1
        test = lbnz.div(fld)

        self.assertAlmostEqualWithMagExceptBoundary(expected, test, 6)

    def test_div3(self):
        # F = y * y, 2 * x * y, 3 * z * z
        # expect that div_F = 2 * x + 6 * z

        fld = lbnz.y * lbnz.y, 2 * lbnz.x * lbnz.y, 3 * lbnz.z * lbnz.z
        expected = 2 * lbnz.x + 6 * lbnz.z
        test = lbnz.div(fld)

        self.assertAlmostEqualWithMagExceptBoundary(expected, test, 6)

    def test_lapacian(self):
        # f = x * y * y + z * z * z
        # expect that grad_f = y * y, 2 * x * y, 3 * z * z

        fld = lbnz.x * lbnz.y * lbnz.y + lbnz.z * lbnz.z * lbnz.z
        e0, e1, e2 = lbnz.y * lbnz.y, 2 * lbnz.x * lbnz.y, 3 * lbnz.z * lbnz.z
        g0, g1, g2 = lbnz.grad(fld)

        self.assertAlmostEqualWithMagExceptBoundary(e0, g0, 3)
        self.assertAlmostEqualWithMagExceptBoundary(e1, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(e2, g2, 4)

        # F = y * y, 2 * x * y, 3 * z * z = grad(f)
        # expect that div_F = 2 * x + 6 * z

        expected = 2 * lbnz.x + 6 * lbnz.z
        test = lbnz.div((g0, g1, g2))

        self.assertAlmostEqualWithMagExceptBoundary(expected, test, 2)

        # f = x * y * y + z * z * z
        # expect that lapacian_f = 2 * x + 6 * z

        fld = lbnz.x * lbnz.y * lbnz.y + lbnz.z * lbnz.z * lbnz.z
        expected = 2 * lbnz.x + 6 * lbnz.z
        test = lbnz.laplacian(fld)

        self.assertAlmostEqualWithMagExceptBoundary(expected, test, 2)

    def test_div_sensibility(self):
        # F = y * y, 2 * x * y, 3 * z * z
        fld = lbnz.y * lbnz.y, 2 * lbnz.x * lbnz.y, 3 * lbnz.z * lbnz.z

        noise = (
            th.randn_like(fld[0]) * th.std(fld[0]) * 1e-6,
            th.randn_like(fld[1]) * th.std(fld[1]) * 1e-6,
            th.randn_like(fld[2]) * th.std(fld[2]) * 1e-6,
        )
        fld_ = fld[0] + noise[0], fld[1] + noise[1], fld[2] + noise[2]
        self.assertAlmostEqualWithMagExceptBoundary(fld[0], fld_[0], 5)
        self.assertAlmostEqualWithMagExceptBoundary(fld[1], fld_[1], 5)
        self.assertAlmostEqualWithMagExceptBoundary(fld[2], fld_[2], 5)

        div1 = lbnz.div(fld)
        div2 = lbnz.div(fld_)
        self.assertAlmostEqualWithMagExceptBoundary(div1, div2, 4)

        diff_div1_0, diff_div1_1, diff_div1_2 = lbnz.diff(div1)
        diff_div2_0, diff_div2_1, diff_div2_2 = lbnz.diff(div2)
        self.assertAlmostEqualWithMagExceptBoundary(diff_div1_0, diff_div2_0, 2)
        self.assertAlmostEqualWithMagExceptBoundary(0.0, diff_div2_1, 2)
        self.assertAlmostEqualWithMagExceptBoundary(diff_div1_2, diff_div2_2, 2)

    def test_curl1(self):
        # F = (y, -x, 0)
        # expect that curl_F = (0, 0, -2)

        fld = lbnz.y, -lbnz.x, lbnz.zero
        expt_x, expt_y, expt_z = lbnz.zero, lbnz.zero, -2 * lbnz.one
        test_x, test_y, test_z = lbnz.curl(fld)

        self.assertAlmostEqualWithMagExceptBoundary(expt_x, test_x, 6)
        self.assertAlmostEqualWithMagExceptBoundary(expt_y, test_y, 6)
        self.assertAlmostEqualWithMagExceptBoundary(expt_z, test_z, 6)

    def test_curl2(self):
        # F = (0, - x**2, 0)
        # expect that curl_F = (0, 0, -2 * x)

        fld = lbnz.zero, -lbnz.x * lbnz.x, lbnz.zero
        expt_x, expt_y, expt_z = lbnz.zero, lbnz.zero, -2 * lbnz.x
        test_x, test_y, test_z = lbnz.curl(fld)

        self.assertAlmostEqualWithMagExceptBoundary(expt_x, test_x, 6)
        self.assertAlmostEqualWithMagExceptBoundary(expt_y, test_y, 6)
        self.assertAlmostEqualWithMagExceptBoundary(expt_z, test_z, 6)

        self.assertAlmostEqualWithMagnitude(expt_x, test_x, 6)
        self.assertAlmostEqualWithMagnitude(expt_y, test_y, 6)
        self.assertAlmostEqualWithMagnitude(expt_z, test_z, 6)

    def test_zero_curl(self):
        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.x))

        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g0, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g2, 6)

        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.y))

        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g0, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g2, 6)

        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.z))

        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g0, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g2, 6)

        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.z * lbnz.z + lbnz.x * lbnz.y))

        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g0, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g1, 6)
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g2, 6)

    def test_zero_div(self):
        g = lbnz.div(lbnz.curl(lbnz.thetaphir.phi))
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g, 6)

        g = lbnz.div(lbnz.curl(lbnz.thetaphir.theta))
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g, 6)

        g = lbnz.div(lbnz.curl(lbnz.thetaphir.r))
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g, 6)

        phx, phy, phz = lbnz.thetaphir.phi
        thx, thy, thz = lbnz.thetaphir.theta
        rx, ry, rz = lbnz.thetaphir.r
        g = lbnz.div(
            lbnz.curl((rx * rx + phx * thx, ry * ry + phy * thy, rz * rz + phz * thz))
        )
        self.assertAlmostEqualWithMagExceptBoundary(lbnz.zero, g, 6)



# -*- coding: utf-8 -*-

import torch as th


_name_ = "phi,rho"
_params_ = ("phi", "rho")


def transform(phi, rho, **kwargs):
    x = rho * th.cos(phi)
    y = rho * th.sin(phi)
    return x, y


def dtransform(phi, rho, dphi, drho, **kwargs):
    dx = -rho * th.sin(phi) * dphi + th.cos(phi) * drho
    dy = -rho * th.cos(phi) * dphi + th.sin(phi) * drho

    return dx, dy


def inverse(x, y, **kwargs):
    rho = th.sqrt(x * x + y * y)
    phi = th.atan2(y, x)
    return phi, rho


def dinverse(x, y, dx, dy, **kwargs):
    rhosq = x * x + y * y
    rho = th.sqrt(rhosq)

    drho = (x * dx + y * dy) / rho
    dphi = (x * dy - y * dx) / rhosq

    return dphi, drho



import torch.nn as nn


class SELayer(nn.Module):
    def __init__(self, channel, reduction=16, conv=None):
        super(SELayer, self).__init__()

        self.avg_pool = None
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction + 1, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction + 1, channel, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x):
        sz = x.size()

        if len(sz) == 3:
            if self.avg_pool is None:
                self.avg_pool = nn.AdaptiveAvgPool1d(1)
            y = self.avg_pool(x)
            y = y.view(sz[0], sz[1])
            y = self.fc(y).view(sz[0], sz[1], 1)
        if len(sz) == 4:
            if self.avg_pool is None:
                self.avg_pool = nn.AdaptiveAvgPool2d(1)
            y = self.avg_pool(x)
            y = y.view(sz[0], sz[1])
            y = self.fc(y).view(sz[0], sz[1], 1, 1)
        if len(sz) == 5:
            if self.avg_pool is None:
                self.avg_pool = nn.AdaptiveAvgPool3d(1)
            y = self.avg_pool(x)
            y = y.view(sz[0], sz[1])
            y = self.fc(y).view(sz[0], sz[1], 1, 1, 1)
        return x * y.expand_as(x)


class SEBasicBlock(nn.Module):
    extension = 1
    least_required_dim = 1

    def __init__(self, dim, step, ix, tx, relu, conv, reduction=16):
        super(SEBasicBlock, self).__init__()
        self.step = step
        self.relu = relu
        self.ix = ix
        self.tx = tx

        self.conv1 = conv(dim, dim, kernel_size=3, stride=1, padding=1)
        self.conv2 = conv(dim, dim, kernel_size=3, stride=1, padding=1)
        self.se = SELayer(dim, reduction)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.se(y)
        y = x + y

        return y


class SEBottleneck(nn.Module):
    extension = 4
    least_required_dim = 1

    def __init__(self, dim, step, ix, tx, relu, conv, reduction=16):
        super(SEBottleneck, self).__init__()
        self.step = step
        self.relu = relu
        self.ix = ix
        self.tx = tx

        self.conv1 = conv(dim, dim // 4, kernel_size=1, bias=False)
        self.conv2 = conv(dim // 4, dim // 4, kernel_size=3, bias=False, padding=1)
        self.conv3 = conv(dim // 4, dim, kernel_size=1, bias=False)
        self.se = SELayer(dim, reduction)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.relu(y)
        y = self.conv3(y)
        y = self.se(y)
        y = x + y

        return y



# -*- coding: utf-8 -*-

import torch as th
import torch.nn as nn

from leibniz.nn.layer.hyperbolic import BasicBlock, Bottleneck


class HyperBasic(nn.Module):
    extension = 1
    least_required_dim = 1

    def __init__(self, dim, step, ix, tx, relu, conv, reduction=16):
        super(HyperBasic, self).__init__()
        self.dim = dim
        self.step = step
        self.ix = ix
        self.tx = tx

        self.input = BasicBlock(dim, 2 * dim, step, relu, conv, reduction=reduction)
        self.output = BasicBlock(4 * dim, dim, step, relu, conv, reduction=reduction)

    def forward(self, x):
        input = self.input(x)
        velo = input[:, : self.dim]
        theta = input[:, self.dim :]

        step = self.step * velo

        y1 = (x + th.tanh(theta)) * th.exp(step * th.sin(theta)) - th.tanh(theta)
        y2 = (x + th.tanh(theta)) * th.exp(-step * th.cos(theta)) - th.tanh(theta)
        y3 = (x - th.tanh(theta)) * th.exp(step * th.sin(theta)) + th.tanh(theta)
        y4 = (x - th.tanh(theta)) * th.exp(-step * th.cos(theta)) + th.tanh(theta)
        ys = th.cat((y1, y2, y3, y4), dim=1)

        return x + self.output(ys)


class HyperBottleneck(nn.Module):
    extension = 4
    least_required_dim = 1

    def __init__(self, dim, step, ix, tx, relu, conv, reduction=16):
        super(HyperBottleneck, self).__init__()
        self.dim = dim
        self.step = step
        self.ix = ix
        self.tx = tx

        self.input = Bottleneck(dim, 2 * dim, step, relu, conv, reduction=reduction)
        self.output = Bottleneck(4 * dim, dim, step, relu, conv, reduction=reduction)

    def forward(self, x):
        input = self.input(x)
        velo = input[:, : self.dim]
        theta = input[:, self.dim :]

        step = self.step * velo

        y1 = (x + th.tanh(theta)) * th.exp(step * th.sin(theta)) - th.tanh(theta)
        y2 = (x + th.tanh(theta)) * th.exp(-step * th.cos(theta)) - th.tanh(theta)
        y3 = (x - th.tanh(theta)) * th.exp(step * th.sin(theta)) + th.tanh(theta)
        y4 = (x - th.tanh(theta)) * th.exp(-step * th.cos(theta)) + th.tanh(theta)
        ys = th.cat((y1, y2, y3, y4), dim=1)

        return x + self.output(ys)



# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch.nn as nn

from leibniz.nn.net.unet import Enconv, Transform, Block
from leibniz.nn.conv import (
    DepthwiseSeparableConv1d,
    DepthwiseSeparableConv2d,
    DepthwiseSeparableConv3d,
)
from leibniz.nn.layer.cbam import CBAM

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ResNet(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        block=None,
        attn=None,
        relu=None,
        layers=4,
        ratio=1,
        dropout_prob=0.5,
        vblks=None,
        factors=None,
        spatial=(256, 256),
        normalizor="batch",
        padding=None,
    ):
        super().__init__()

        extension = block.extension

        spatial = np.array(spatial, dtype=np.int32)
        dim = len(spatial)
        self.dim = dim
        Conv = self.get_conv_for_prepare()
        TConv = self.get_conv_for_transform()

        ratio = np.exp2(ratio)
        factors = np.array(factors)
        factors = np.exp2(factors)
        num_filters = int(in_channels * ratio)

        self.ratio = ratio
        self.vblks = vblks
        self.factors = factors
        logger.info("---------------------------------------")
        logger.info("ratio: %f", ratio)
        logger.info("vblks: [%s]", ", ".join(map(str, vblks)))
        logger.info("factors: [%s]", ", ".join(map(str, factors)))
        logger.info("---------------------------------------")

        self.layers = layers
        self.in_channels = in_channels
        self.num_filters = num_filters
        self.out_channels = out_channels

        if relu is None:
            relu = nn.ReLU(inplace=True)

        if attn is None:
            attn = CBAM

        ex = extension
        c0 = int(ex * num_filters)
        # accumulated_factors = np.cumprod(factors, axis=0)[-1]
        # ch = int(c0 * accumulated_factors)

        self.conv_padding = 1
        self.iconv = Conv(in_channels, c0, kernel_size=7, padding=3, groups=1)
        self.enconvs = nn.ModuleList()
        self.transforms = nn.ModuleList()

        self.spatial = [np.array(spatial, dtype=np.int32)]
        self.channel_sizes = [np.array(c0, dtype=np.int32)]
        for ix in range(layers):
            least_factor = ex
            factor = factors[ix]
            self.channel_sizes.append(
                np.array(
                    self.channel_sizes[ix] * factor // least_factor * least_factor,
                    dtype=np.int32,
                )
            )

            ci, co = self.channel_sizes[ix].item(), self.channel_sizes[ix + 1].item()
            logger.info("%d - ci: %d, co: %d", ix, ci, co)

            self.enconvs.append(
                Block(
                    Enconv(ci, co, size=spatial, conv=TConv, padding=padding),
                    activation=True,
                    dropout=dropout_prob,
                    relu=relu,
                    attn=attn,
                    dim=self.dim,
                    normalizor=normalizor,
                    conv=TConv,
                )
            )
            self.transforms.append(
                Transform(co, co, nblks=vblks[ix], block=block, relu=relu, conv=TConv)
            )

        self.oconv = Conv(co, out_channels, kernel_size=3, padding=1, groups=1)

    def get_conv_for_prepare(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def get_conv_for_transform(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def forward(self, x):
        # sz = x.size()

        dnt = self.iconv(x)
        for ix in range(self.layers):
            dnt, enc = self.transforms[ix](self.enconvs[ix](dnt))

        return self.oconv(dnt)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="leibniz",
    version="0.2.0",
    author="Mingli Yuan",
    author_email="mingli.yuan@gmail.com",
    description="Leibniz: a package providing facilities to express learnable differential equations based on PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mountain/leibniz",
    project_urls={
        "Documentation": "https://github.com/mountain/leibniz",
        "Source": "https://github.com/mountain/leibniz",
        "Tracker": "https://github.com/mountain/leibniz/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "cached_property",
        "torchpwl",
        "torch",
        "numpy",
    ],
    test_suite="pytest",
    tests_require=["pytest"],
)



# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch as th
import torch.nn as nn

from leibniz.nn.conv import (
    DepthwiseSeparableConv1d,
    DepthwiseSeparableConv2d,
    DepthwiseSeparableConv3d,
)
from leibniz.nn.layer.cbam import CBAM
from leibniz.nn.net.hyptube import HypTube

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class Enconv(nn.Module):
    def __init__(
        self, in_channels, out_channels, size=(256, 256), conv=nn.Conv2d, padding=None
    ):

        super(Enconv, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.size = size

        if len(size) == 1:
            self.scale = nn.Upsample(size=tuple(size), mode="linear")
        elif len(size) == 2:
            self.scale = nn.Upsample(size=tuple(size), mode="bilinear")
        elif len(size) == 3:
            self.scale = nn.Upsample(size=tuple(size), mode="trilinear")

        self.padding = padding
        if padding is not None:
            self.conv = conv(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=0,
                dilation=1,
                groups=1,
            )
        else:
            self.conv = conv(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                dilation=1,
                groups=1,
            )

    def forward(self, x):
        ratio = (np.array(x.size())[-len(self.size) :].prod()) / (
            np.array(self.size).prod()
        )
        if ratio < 1.0:
            x = self.scale(x)
            if self.padding is not None:
                x = self.padding(x)
            x = self.conv(x)
        else:
            if self.padding is not None:
                x = self.padding(x)
            x = self.conv(x)
            x = self.scale(x)

        return x


class Deconv(nn.Module):
    def __init__(
        self, in_channels, out_channels, size=(256, 256), conv=nn.Conv2d, padding=None
    ):
        super(Deconv, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.size = size

        if len(size) == 1:
            self.scale = nn.Upsample(size=tuple(size), mode="linear")
        elif len(size) == 2:
            self.scale = nn.Upsample(size=tuple(size), mode="bilinear")
        elif len(size) == 3:
            self.scale = nn.Upsample(size=tuple(size), mode="trilinear")

        self.padding = padding
        if padding is not None:
            self.conv = conv(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=0,
                dilation=1,
                groups=1,
            )
        else:
            self.conv = conv(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                dilation=1,
                groups=1,
            )

    def forward(self, x):
        ratio = (np.array(x.size())[-len(self.size) :].prod()) / (
            np.array(self.size).prod()
        )
        if ratio < 1.0:
            x = self.scale(x)
            if self.padding is not None:
                x = self.padding(x)
            x = self.conv(x)
        else:
            if self.padding is not None:
                x = self.padding(x)
            x = self.conv(x)
            x = self.scale(x)

        return x


class Transform(nn.Module):
    def __init__(
        self, in_channels, out_channels, nblks=0, block=None, relu=None, conv=nn.Conv2d
    ):

        super(Transform, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        if relu is None:
            relu = nn.ReLU(inplace=True)

        self.blocks = None
        if nblks > 0 and block is not None:
            blocks = []
            for i in range(1, nblks - 1, 1):
                blocks.append(
                    block(
                        self.out_channels,
                        step=1.0 / nblks,
                        ix=i,
                        tx=nblks,
                        relu=relu,
                        conv=conv,
                    )
                )
                blocks.append(relu)
            blocks.append(
                block(
                    self.out_channels,
                    step=1.0 / nblks,
                    ix=nblks,
                    tx=nblks,
                    relu=relu,
                    conv=conv,
                )
            )
            self.blocks = nn.Sequential(*blocks)

    def forward(self, x):

        if self.blocks is not None:
            return self.blocks(x), x
        else:
            return x, x


class Block(nn.Module):
    def __init__(
        self,
        transform,
        activation=True,
        dropout=-1,
        relu=None,
        attn=CBAM,
        dim=2,
        normalizor="batch",
        conv=None,
    ):

        super(Block, self).__init__()
        self.activation = activation
        self.dropout_flag = dropout > 0 and self.training
        self.blocks = None

        self.transform = transform

        self.attn = attn(transform.out_channels, conv=conv)

        if self.activation:
            if relu is not None:
                self.lrelu = relu
            else:
                self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

        self.normalizor = None
        if normalizor == "batch":
            if dim == 1:
                self.normalizor = nn.BatchNorm1d(transform.out_channels, affine=True)
            elif dim == 2:
                self.normalizor = nn.BatchNorm2d(transform.out_channels, affine=True)
            elif dim == 3:
                self.normalizor = nn.BatchNorm3d(transform.out_channels, affine=True)

        elif normalizor == "instance":
            if dim == 1:
                self.normalizor = nn.InstanceNorm1d(transform.out_channels)
            elif dim == 2:
                self.normalizor = nn.InstanceNorm2d(transform.out_channels)
            elif dim == 3:
                self.normalizor = nn.InstanceNorm3d(transform.out_channels)

        elif normalizor == "layer":
            self.normalizor = nn.LayerNorm(
                tuple([transform.out_channels]) + tuple(transform.size)
            )

        if self.dropout_flag:
            if dim == 1:
                self.drop = nn.Dropout(p=dropout, inplace=True)
            elif dim == 2:
                self.drop = nn.Dropout2d(p=dropout, inplace=True)
            elif dim == 3:
                self.drop = nn.Dropout3d(p=dropout, inplace=True)

    def forward(self, *xs):

        x = th.cat(xs, dim=1)

        if self.activation:
            x = self.lrelu(x)

        x = self.transform(x)

        if self.normalizor:
            x = self.normalizor(x)

        if self.dropout_flag and self.training:
            x = self.drop(x)

        x = self.attn(x)

        return x


class UNet(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        block=None,
        attn=None,
        relu=None,
        layers=4,
        ratio=2,
        enhencer=None,
        ksize_in=7,
        dropout_prob=0.1,
        vblks=None,
        hblks=None,
        scales=None,
        factors=None,
        spatial=(256, 256),
        normalizor="batch",
        padding=None,
        final_normalized=True,
    ):
        super().__init__()

        extension = block.extension
        lrd = block.least_required_dim

        spatial = np.array(spatial, dtype=np.int32)
        dim = len(spatial)
        self.dim = dim
        Conv = self.get_conv_for_prepare()
        TConv = self.get_conv_for_transform()

        scales = np.array(scales)
        if scales.shape[0] != layers:
            raise ValueError("scales should have %d layers at dim 0!" % layers)
        if len(scales.shape) == 1:
            scales = scales.reshape([layers, 1])
        if len(scales.shape) != 2:
            raise ValueError(
                "scales should have length 2 to be compatible with spatial dimensions!"
            )

        ratio = np.exp2(ratio)
        factors = np.array(factors + [0.0])
        scales = np.exp2(scales)
        factors = np.exp2(factors)
        num_filters = int(in_channels * ratio)

        self.final_normalized = final_normalized
        self.ratio = ratio
        self.hblks = hblks
        self.vblks = vblks
        self.scales = scales
        self.factors = factors
        logger.info("---------------------------------------")
        logger.info("ratio: %f", ratio)
        logger.info("vblks: [%s]", ", ".join(map(str, vblks)))
        logger.info("hblks: [%s]", ", ".join(map(str, hblks)))
        logger.info("scales: [%s]", ", ".join(map(str, scales)))
        logger.info("factors: [%s]", ", ".join(map(str, factors[0:4])))
        logger.info("---------------------------------------")

        self.exceeded = np.any(np.cumprod(scales, axis=0) * spatial < 1) or np.any(
            (in_channels * ratio * np.cumprod(factors)) < lrd
        )
        if not self.exceeded:
            self.layers = layers
            self.in_channels = in_channels
            self.num_filters = num_filters
            self.out_channels = out_channels

            if enhencer is None:
                if self.dim == 2:
                    enhencer = HypTube
            self.enhencer_in = None
            self.enhencer_out = None
            self.enhencer_mid = None

            if relu is None:
                relu = nn.ReLU(inplace=True)

            if attn is None:
                attn = CBAM

            ex = extension
            c0 = int(ex * num_filters)
            if padding:
                self.conv_padding = 0
                self.iconv = nn.Sequential(
                    padding,
                    Conv(
                        in_channels,
                        c0,
                        kernel_size=ksize_in,
                        padding=(ksize_in - 1) // 2,
                        groups=1,
                    ),
                )
                self.oconv = nn.Sequential(
                    padding,
                    Conv(
                        c0,
                        out_channels,
                        kernel_size=3,
                        padding=self.conv_padding,
                        bias=False,
                        groups=1,
                    ),
                )
            else:
                self.conv_padding = 1
                self.iconv = Conv(
                    in_channels,
                    c0,
                    kernel_size=ksize_in,
                    padding=(ksize_in - 1) // 2,
                    groups=1,
                )
                self.oconv = Conv(
                    c0,
                    out_channels,
                    kernel_size=3,
                    padding=self.conv_padding,
                    bias=False,
                    groups=1,
                )

            if final_normalized:
                self.relu6 = nn.ReLU6()

            self.enconvs = nn.ModuleList()
            self.dnforms = nn.ModuleList()
            self.hzforms = nn.ModuleList()
            self.upforms = nn.ModuleList()
            self.deconvs = nn.ModuleList()

            self.spatial = [np.array(spatial, dtype=np.int32)]
            self.channel_sizes = [np.array(c0, dtype=np.int32)]
            for ix in range(layers):
                least_factor = ex
                scale, factor = scales[ix], factors[ix]
                self.spatial.append(np.array(self.spatial[ix] * scale, dtype=np.int32))
                self.channel_sizes.append(
                    np.array(
                        self.channel_sizes[ix] * factor // least_factor * least_factor,
                        dtype=np.int32,
                    )
                )

                ci, co = (
                    self.channel_sizes[ix].item(),
                    self.channel_sizes[ix + 1].item(),
                )
                szi, szo = self.spatial[ix + 1], self.spatial[ix]
                logger.info("%d - ci: %d, co: %d", ix, ci, co)
                logger.info(
                    "%d - szi: [%s], szo: [%s]",
                    ix,
                    ", ".join(map(str, szi)),
                    ", ".join(map(str, szo)),
                )

                self.exceeded = (
                    self.exceeded
                    or ci < lrd
                    or co < lrd
                    or szi.min() < 1
                    or szo.min() < 1
                )
                if not self.exceeded:
                    try:
                        dropout_flag = (layers - ix) * 3 < layers
                        dropout = dropout_prob if dropout_flag else -1
                        self.enconvs.append(
                            Block(
                                Enconv(ci, co, size=szi, conv=TConv, padding=padding),
                                activation=True,
                                dropout=dropout,
                                relu=relu,
                                attn=attn,
                                dim=self.dim,
                                normalizor=normalizor,
                                conv=TConv,
                            )
                        )
                        self.dnforms.append(
                            Transform(
                                co,
                                co,
                                nblks=vblks[ix],
                                block=block,
                                relu=relu,
                                conv=TConv,
                            )
                        )
                        self.hzforms.append(
                            Transform(
                                co,
                                co,
                                nblks=hblks[ix],
                                block=block,
                                relu=relu,
                                conv=TConv,
                            )
                        )
                        self.deconvs.append(
                            Block(
                                Deconv(
                                    co * 2, ci, size=szo, conv=TConv, padding=padding
                                ),
                                activation=True,
                                dropout=False,
                                relu=relu,
                                attn=attn,
                                dim=self.dim,
                                normalizor=normalizor,
                                conv=TConv,
                            )
                        )
                        self.upforms.append(
                            Transform(
                                ci,
                                ci,
                                nblks=vblks[ix],
                                block=block,
                                relu=relu,
                                conv=TConv,
                            )
                        )
                    except Exception as e:
                        logger.exception(e)
                        self.exceeded = True
                else:
                    logger.error("scales are exceeded!")
                    raise ValueError("scales exceeded!")

            if self.dim == 2 and enhencer is not None:
                self.enhencer_mid = enhencer(co, (c0 + 1) // 2, co)

    def get_conv_for_prepare(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def get_conv_for_transform(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def forward(self, x):
        if self.exceeded:
            raise ValueError("scales exceeded!")

        dnt = self.iconv(x)
        if self.enhencer_in is not None:
            dnt = self.enhencer_in(dnt)

        hzts = []
        for ix in range(self.layers):
            dnt, enc = self.dnforms[ix](self.enconvs[ix](dnt))
            hzt, _ = self.hzforms[ix](enc)
            hzts.append(hzt)

        if self.enhencer_mid is None:
            upt = dnt
        else:
            upt = self.enhencer_mid(dnt)

        for ix in range(self.layers - 1, -1, -1):
            hzt = hzts[ix]
            upt, dec = self.upforms[ix](self.deconvs[ix](upt, hzt))

        if self.enhencer_out is not None:
            upt = self.enhencer_out(upt)

        if self.final_normalized:
            return self.relu6(self.oconv(upt)) / 6
        else:
            return self.oconv(upt)



# -*- coding: utf-8 -*-

import torch as th
from torch import nn


class Swish(th.nn.Module):
    def __init__(self):
        super(Swish, self).__init__()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        return x * self.sigmoid(x)


class Mish(th.nn.Module):
    def __init__(self):
        super(Mish, self).__init__()
        self.tanh = nn.Tanh()
        self.softplus = nn.Softplus()

    def forward(self, x):
        return x * self.tanh(self.softplus(x))


class CappingRelu(th.nn.Module):
    def __init__(self):
        super(CappingRelu, self).__init__()
        self.leaky = nn.LeakyReLU(inplace=True)

    def forward(self, x):
        if x.dtype == th.cfloat or x.dtype == th.cdouble:
            return th.clamp(self.leaky(x.real), max=10) + 1j * th.clamp(
                self.leaky(x.imag), max=10
            )
        return th.clamp(self.leaky(x), max=10)


class Sigmoid(th.nn.Module):
    def __init__(self):
        super(Sigmoid, self).__init__()

    def forward(self, x):
        return (1 + th.tanh(x)) / 2.0


class Atanh(nn.Module):
    def __init__(self):
        super(Atanh, self).__init__()

    def forward(self, x):
        return th.log1p(2 * x / (1 - x)) / 2


class ComplexReLU(nn.Module):
    def __init__(self):
        super(ComplexReLU, self).__init__()
        self.relu = nn.ReLU(inplace=True)

    def forward(self, input):
        return self.relu(input.real) + 1j * self.relu(input.imag)


class ComplexLinear(nn.Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(ComplexLinear, self).__init__()
        self.rfc = nn.Linear(in_channels, out_channels, bias)
        self.ifc = nn.Linear(in_channels, out_channels, bias)

    def forward(self, input):
        return (
            self.rfc(input.real)
            - self.ifc(input.imag)
            + 1j * (self.rfc(input.imag) + self.ifc(input.real))
        )



# -*- coding: utf-8 -*-

import torch as th
import leibniz as lbnz

from cached_property import cached_property
from torch import Tensor
from typing import Tuple


class Elements:
    def __init__(self, **kwargs):
        self.default_device = -1

        if lbnz._grid_.basis == "theta,phi,r":
            self.dL1 = lbnz.r * th.cos(lbnz.phi) * lbnz.dtheta
            self.dL2 = lbnz.r * lbnz.dphi
            self.dL3 = lbnz.dr
        if lbnz._grid_.basis == "x,y,z":
            self.dL1 = lbnz.dx
            self.dL2 = lbnz.dy
            self.dL3 = lbnz.dz
        if lbnz._grid_.basis == "lng,lat,alt":
            lbnz.use("theta,phi,r", **kwargs)
            self.dL1 = lbnz.r * th.cos(lbnz.phi) * lbnz.dtheta
            self.dL2 = lbnz.r * lbnz.dphi
            self.dL3 = lbnz.dr

        self.dS3 = self.dL1 * self.dL2
        self.dS1 = self.dL3 * self.dL2
        self.dS2 = self.dL3 * self.dL1

        self.dVol = self.dL1 * self.dL2 * self.dL3

    def get_device(self):
        return self.default_device

    def set_device(self, ix):
        self.dL1 = self.dL1.cuda(device=ix)
        self.dL2 = self.dL2.cuda(device=ix)
        self.dL3 = self.dL3.cuda(device=ix)
        self.dS1 = self.dS1.cuda(device=ix)
        self.dS2 = self.dS2.cuda(device=ix)
        self.dS3 = self.dS3.cuda(device=ix)
        self.dVol = self.dVol.cuda(device=ix)
        self.default_device = ix

    @cached_property
    def dL(self) -> Tuple[Tensor, Tensor, Tensor]:
        return self.dL1, self.dL2, self.dL3

    @cached_property
    def dS(self) -> Tuple[Tensor, Tensor, Tensor]:
        return self.dS1, self.dS2, self.dS3

    @cached_property
    def dV(self) -> Tensor:
        return self.dVol



# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch as th
import torch.nn as nn

from leibniz.nn.conv import (
    DepthwiseSeparableConv1d,
    DepthwiseSeparableConv2d,
    DepthwiseSeparableConv3d,
)
from leibniz.nn.layer.hyperbolic import Bottleneck

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ResNetZ(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        layers=4,
        ratio=1,
        spatial=(256, 256),
        conv_class=None,
        relu=None,
        normalizor=None,
    ):
        super().__init__()

        spatial = np.array(spatial, dtype=np.int32)
        dim = len(spatial)
        self.dim = dim
        self.ratio = np.power(2, ratio)
        self.layers = layers
        self.in_channels = int(in_channels)
        self.num_filters = int(in_channels * self.ratio)
        self.out_channels = int(out_channels)
        self.spatial = [np.array(spatial, dtype=np.int32)]

        logger.info("---------------------------------------")
        logger.info("dim: %f", self.dim)
        logger.info("ratio: %f", self.ratio)
        logger.info("layers: %f", self.layers)
        logger.info("in_channels: %f", self.in_channels)
        logger.info("out_channels: %f", self.out_channels)
        logger.info("num_filters: %f", self.num_filters)
        logger.info("normalizor: %s", normalizor)
        logger.info("---------------------------------------")

        if dim == 1:
            self.bn = nn.BatchNorm1d(self.num_filters, affine=True)
        elif dim == 2:
            self.bn = nn.BatchNorm2d(self.num_filters, affine=True)
        elif dim == 3:
            self.bn = nn.BatchNorm3d(self.num_filters, affine=True)

        if conv_class is None:
            self.conv_class = self.get_conv_class()
        else:
            self.conv_class = conv_class
        if relu is None:
            self.relu = nn.ReLU(inplace=True)
        else:
            self.relu = relu

        if normalizor == "relu6":
            self.normalizor = nn.ReLU6()
            self.scale = 1.0 / 6.0
            self.bias = 0.0
        elif normalizor == "sigmoid":
            self.normalizor = nn.Sigmoid()
            self.scale = 1.0
            self.bias = 0.0
        elif normalizor == "tanh":
            self.normalizor = nn.Tanh()
            self.scale = 1.0
            self.bias = 0.0
        elif normalizor == "softmax":
            self.normalizor = nn.Softmax()
            self.scale = 1.0
            self.bias = 0.0
        else:
            self.normalizor = None
            self.scale = 1.0
            self.bias = 0.0

        self.iconv = self.conv_class(
            self.in_channels, self.num_filters, kernel_size=7, padding=3, groups=1
        )
        self.oconv = self.conv_class(
            self.num_filters,
            self.out_channels,
            kernel_size=3,
            padding=1,
            groups=1,
            bias=False,
        )

        step_length = 1.0 / self.layers
        self.order1 = Bottleneck(
            self.num_filters,
            2 * self.num_filters,
            step_length,
            self.relu,
            self.conv_class,
            reduction=16,
        )
        self.order2 = Bottleneck(
            4 * self.num_filters + 1,
            2 * self.num_filters,
            step_length,
            self.relu,
            self.conv_class,
            reduction=16,
        )
        self.order3 = Bottleneck(
            7 * self.num_filters + 1,
            2 * self.num_filters,
            step_length,
            self.relu,
            self.conv_class,
            reduction=16,
        )

    def get_conv_class(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def forward(self, x):
        x0 = self.bn(self.iconv(x))
        rslt = self.order1(x0)
        velo = rslt[:, : self.num_filters]
        theta = rslt[:, self.num_filters :]
        du0 = velo * th.cos(theta)
        dv0 = velo * th.sin(theta)

        for _ in range(self.layers):
            x1 = x0 * (1 + dv0 / self.layers) + du0 / self.layers
            x1 = self.relu(x1)

            dd = self.order2(
                th.cat(
                    [x0, x1, du0, dv0, th.ones_like(x0[:, 0:1]) * _ / self.layers],
                    dim=1,
                )
            )
            du1 = dd[:, self.num_filters * 0 : self.num_filters * 1]
            dv1 = dd[:, self.num_filters * 1 : self.num_filters * 2]

            x2 = x1 * (1 + dv1 / self.layers) + du1 / self.layers
            x2 = self.relu(x2)

            dd = self.order3(
                th.cat(
                    [
                        x0,
                        x1,
                        x2,
                        du0,
                        dv0,
                        du1,
                        dv1,
                        th.ones_like(x1[:, 0:1]) * _ / self.layers,
                    ],
                    dim=1,
                )
            )
            du2 = dd[:, self.num_filters * 0 : self.num_filters * 1]
            dv2 = dd[:, self.num_filters * 1 : self.num_filters * 2]

            x3 = x2 * (1 + dv2 / self.layers) + du2 / self.layers
            x3 = self.relu(x3)

        out = self.oconv(x3)
        if self.normalizor:
            return self.normalizor(out) * self.scale + self.bias
        else:
            return out * self.scale + self.bias



from .tsit5 import Tsit5Solver
from .dopri5 import Dopri5Solver
from .fixed_grid import Euler, Midpoint, RK4
from .fixed_adams import AdamsBashforth, AdamsBashforthMoulton
from .adams import VariableCoefficientAdamsBashforth
from .misc import _check_inputs

SOLVERS = {
    "explicit_adams": AdamsBashforth,
    "fixed_adams": AdamsBashforthMoulton,
    "adams": VariableCoefficientAdamsBashforth,
    "tsit5": Tsit5Solver,
    "dopri5": Dopri5Solver,
    "euler": Euler,
    "midpoint": Midpoint,
    "rk4": RK4,
}


def odeint(func, y0, t, rtol=1e-7, atol=1e-9, method=None, options=None):
    """Integrate a system of ordinary differential equations.

    Solves the initial value problem for a non-stiff system of first order ODEs:
        ```
        dy/dt = func(t, y), y(t[0]) = y0
        ```
    where y is a Tensor of any shape.

    Output dtypes and numerical precision are based on the dtypes of the inputs `y0`.

    Args:
        func: Function that maps a Tensor holding the state `y` and a scalar Tensor
            `t` into a Tensor of state derivatives with respect to time.
        y0: N-D Tensor giving starting value of `y` at time point `t[0]`. May
            have any floating point or complex dtype.
        t: 1-D Tensor holding a sequence of time points for which to solve for
            `y`. The initial time point should be the first element of this sequence,
            and each time must be larger than the previous time. May have any floating
            point dtype. Converted to a Tensor with float64 dtype.
        rtol: optional float64 Tensor specifying an upper bound on relative error,
            per element of `y`.
        atol: optional float64 Tensor specifying an upper bound on absolute error,
            per element of `y`.
        method: optional string indicating the integration method to use.
        options: optional dict of configuring options for the indicated integration
            method. Can only be provided if a `method` is explicitly set.
        name: Optional name for this operation.

    Returns:
        y: Tensor, where the first dimension corresponds to different
            time points. Contains the solved value of y for each desired time point in
            `t`, with the initial value `y0` being the first element along the first
            dimension.

    Raises:
        ValueError: if an invalid `method` is provided.
        TypeError: if `options` is supplied without `method`, or if `t` or `y0` has
            an invalid dtype.
    """

    tensor_input, func, y0, t = _check_inputs(func, y0, t)

    if options is None:
        options = {}
    elif method is None:
        raise ValueError("cannot supply `options` without specifying `method`")

    if method is None:
        method = "dopri5"

    solver = SOLVERS[method](func, y0, rtol=rtol, atol=atol, **options)
    solution = solver.integrate(t)

    if tensor_input:
        solution = solution[0]
    return solution
